/*
 * linux/sound/arm/ep93xx-alsa-pcm.c -- ALSA PCM interface for the edb93xx ac97 audio
 *
 * Author:      Shrek Wu
 * Created:     July 19, 2005
 * Copyright:   Cirrus Logic, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/device.h>
#include <linux/slab.h>
#include <linux/dma-mapping.h>

#include <linux/version.h>
#include <linux/soundcard.h>


#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/ac97_codec.h>
#include <sound/initval.h>
                                                                                                                             
#include <asm/irq.h>
#include <asm/semaphore.h>
#include <asm/hardware.h>
#include <asm/io.h>
//#include <asm/arch/dma-ep93xx.h>
#include <asm/arch/dma.h>
#include <asm/arch/regs_ac97.h>
#include "ep93xx-alsa-pcm.h"

/*-----------------------------------------------------------------------------------------------*/
#define Alsa_DMA

#ifdef ALSA_AC97_DEBUG
#define DPRINTK( x... )  printk( ##x )
#else
#define DPRINTK( x... )
#endif

/* Mostly just prints what ioctls got called */
/* #define DEBUG 1 */
#ifdef ALSA_AC97_DEBUG
#define DPRINTK_IOCTL( x... )  printk( ##x )
#else
#define DPRINTK_IOCTL( x... )
#endif


static const snd_pcm_hardware_t ep93xx_ac97_pcm_hardware = {
	.info				= 	(SNDRV_PCM_INFO_MMAP |
				  			SNDRV_PCM_INFO_MMAP_VALID |
				  			SNDRV_PCM_INFO_INTERLEAVED |
				  			SNDRV_PCM_INFO_PAUSE),
	/*.formats			= 	SNDRV_PCM_FMTBIT_U8 SNDRV_PCM_FMTBIT_S16_LE,*/
	.formats =		SNDRV_PCM_FMTBIT_U8 | SNDRV_PCM_FMTBIT_S8 |
				SNDRV_PCM_FMTBIT_U16_LE | SNDRV_PCM_FMTBIT_S16_LE |
				SNDRV_PCM_FMTBIT_U16_BE | SNDRV_PCM_FMTBIT_S16_BE |
				SNDRV_PCM_FMTBIT_U32_LE | SNDRV_PCM_FMTBIT_S32_LE |
				SNDRV_PCM_FMTBIT_U32_BE | SNDRV_PCM_FMTBIT_S32_BE,
	.rates				=	(SNDRV_PCM_RATE_8000 | SNDRV_PCM_RATE_11025 |
							SNDRV_PCM_RATE_16000 | SNDRV_PCM_RATE_22050 |
							SNDRV_PCM_RATE_32000 | SNDRV_PCM_RATE_44100 |
							SNDRV_PCM_RATE_48000),
	.rate_min			= 8000,
	.rate_max			= 48000,
	.channels_min		= 1,
	.channels_max		= 2,					
	.period_bytes_min	= 16,
	.period_bytes_max	= 16384,/*32768*/
	.periods_min		= 8,//2,
	.periods_max		= 16,//8,//4,
	.buffer_bytes_max	= 128 * 1024,
	.fifo_size			= 32,
};


/*-------------------------------------------------------------------------------------------*/
#define AUDIO_NAME		"ep93xx-audio"

#define AUDIO_NBFRAGS_DEFAULT	8
#define AUDIO_FRAGSIZE_DEFAULT	16384

/*
 * Translates to:
 * 		s->buf_idx++;
 * 		s->buf_idx %= s->nbfrags;
 *		s->buf = s->buffers + s->buf_idx;
 *
 * So s->buf always points to the s->buf_idx-nth element of s->buffers.
 */
#define NEXT_BUF(_s_,_b_) { \
	(_s_)->_b_##_idx++; \
	(_s_)->_b_##_idx %= (_s_)->nbfrags; \
	(_s_)->_b_ = (_s_)->buffers + (_s_)->_b_##_idx; }

#define AUDIO_ACTIVE(state)	((state)->rd_ref || (state)->wr_ref)


struct semaphore  sLoopsem;

int stoptmp;
unsigned int ReadEnd,WriteEnd;
unsigned int Stack_buf_id[100];

/*-----------------------------------------------------------------------------------------*/

static void print_audio_format( long format )
{
	switch( format )
	{
		case SNDRV_PCM_FORMAT_U8:		   
			DPRINTK( "AFMT_U8\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S8:
			DPRINTK( "AFMT_S8\n" );		   
			break;

		case /*AFMT_S32_BLOCKED*/SNDRV_PCM_FORMAT_S32_LE:
			DPRINTK( "AFMT_S32_BLOCKED\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S16_LE:
			DPRINTK( "AFMT_S16_LE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_S16_BE:
			DPRINTK( "AFMT_S16_BE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_U16_LE:
			DPRINTK( "AFMT_U16_LE\n" );		   
			break;

		case SNDRV_PCM_FORMAT_U16_BE:
		default:
			DPRINTK( "AFMT_U16_BE\n" );		   
			break;
	}
}

/*
 * We convert to 24 bit samples that occupy 32 bits each.
 * Formats we support:
 *
 * AFMT_U8		   
 * AFMT_S16_LE	   	Little endian signed 16
 * AFMT_S8		   
 * AFMT_U16_LE	   	Little endian U16
 * AFMT_S32_BLOCKED	32 bit little endian format, taken from the rme96xx driver.
 *
 */
static void audio_set_format( audio_stream_t * s, long val )
{

	switch( val )
	{
		case /*AFMT_U8*/ SNDRV_PCM_FORMAT_U8:		   
			s->audio_format = val;
			s->audio_stream_bitwidth = 8;
			break;

		case /*AFMT_S8*/ SNDRV_PCM_FORMAT_S8:
			s->audio_format = val;
			s->audio_stream_bitwidth = 8;
			break;

		case /*AFMT_S32_BLOCKED*/ SNDRV_PCM_FORMAT_S32_BE:
		case SNDRV_PCM_FMTBIT_S32_LE:
		
			s->audio_format = SNDRV_PCM_FORMAT_S32_LE;
			s->audio_stream_bitwidth = 32;
			break;

		case /*AFMT_S16_LE*/ SNDRV_PCM_FORMAT_S16_LE:
		case /*AFMT_S16_BE*/ SNDRV_PCM_FORMAT_S16_BE:
			s->audio_format = SNDRV_PCM_FORMAT_S16_LE /*AFMT_S16_LE*/;
			s->audio_stream_bitwidth = 16;
			break;

		case /*AFMT_U16_LE*/ SNDRV_PCM_FORMAT_U16_LE:
		case /*AFMT_U16_BE*/ SNDRV_PCM_FORMAT_U16_BE:
		default:
			s->audio_format = SNDRV_PCM_FORMAT_U16_LE/*AFMT_U16_LE*/;
			s->audio_stream_bitwidth = 16;
			break;
	}

	DPRINTK( "audio_set_format EXIT format set to be (%d) ", 
				(int)s->audio_format );
	print_audio_format( (long)s->audio_format );
}

static  unsigned long copy_to_user_U16_LE_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	unsigned short * user_ptr = (unsigned short *)to; /* 16 bit user buffer */
	short * dma_ptr = (short *)from;
	short right, left;
	
	/*
	 * Compact mode - left is the lower 16 bits.  right is the upper 16 bits.
	 */

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			right = *dma_ptr++;
			left  = *dma_ptr++;
			__put_user( left ^ 0x8000, user_ptr++ );
			__put_user( right^ 0x8000, user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			dma_ptr++;	/* skip right channel sample */
			left  = *dma_ptr++;
			__put_user( left ^ 0x8000, user_ptr++ );
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_U16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	short * user_ptr = (short *)to;	/* 16 bit user buffer */
	int * dma_ptr = (int *)from;	/* 32 bit dma buffer  */

	int shift = state->DAC_bit_width - 16;
	
	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( ((short)( (*dma_ptr++) >> shift )) ^ 0x8000, user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_S16_LE_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	short * user_ptr = (short *)to;	  /* 16 bit user buffer */
	short * dma_ptr = (short *)from;
	short left, right;

	/*
	 * Compact mode - left is the lower 16 bits.  right is the upper 16 bits.
	 */

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			right = *dma_ptr++;
			left  = *dma_ptr++;
			__put_user( left,  user_ptr++ );
			__put_user( right, user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			dma_ptr++; /* skip right sample */
			left  = *dma_ptr++;
			__put_user( left,  user_ptr++ );
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_S16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	short * user_ptr = (short *)to;	/* 16 bit user buffer */
	int * dma_ptr = (int *)from;	/* 32 bit dma buffer  */

	int shift = state->DAC_bit_width - 16;
	
	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( (short)( (*dma_ptr++) >> shift ), user_ptr++ );
			__put_user( (short)( (*dma_ptr++) >> shift ), user_ptr++ );
			to_count -= 4;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( (short)((*dma_ptr++) >> shift), user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count -= 2;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_S8_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  	/*  8 bit user buffer */
	unsigned short * dma_ptr = (unsigned short *)from;    /* 16 bit dma buffer  */
	char right, left;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			right = ((char)(*dma_ptr++ >> 8));
			left  = ((char)(*dma_ptr++ >> 8));
			__put_user( left,  user_ptr++ );
			__put_user( right, user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			dma_ptr++;	/* skip right channel sample */
			left  = ((char)(*dma_ptr++ >> 8));
			__put_user( left,  user_ptr++ );
			to_count--;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_S8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  /*  8 bit user buffer */
	int * dma_ptr = (int *)from;    /* 32 bit dma buffer  */

	int shift = state->DAC_bit_width- 8;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( (char)( (*dma_ptr++) >> shift ), user_ptr++ );
			__put_user( (char)( (*dma_ptr++) >> shift ), user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( (char)((*dma_ptr++) >> shift), user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count--;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_U8_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  	/*  8 bit user buffer */
	unsigned short * dma_ptr = (unsigned short *)from;    /* 16 bit dma buffer  */
	char right, left;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			right = ((char)(*dma_ptr++ >> 8)) ^ 0x80;
			left  = ((char)(*dma_ptr++ >> 8)) ^ 0x80;
			__put_user( left,  user_ptr++ );
			__put_user( right, user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			dma_ptr++;	/* skip right channel sample */
			left  = ((char)(*dma_ptr++ >> 8)) ^ 0x80;
			__put_user( left,  user_ptr++ );
			to_count--;
		}
	}
	
	return total_to_count;
}

static  unsigned long copy_to_user_U8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	unsigned long to_count
)
{
	int total_to_count = to_count;
	char * user_ptr = (char *)to;  /*  8 bit user buffer */
	int * dma_ptr = (int *)from;    /* 32 bit dma buffer  */

	int shift = state->DAC_bit_width - 8;

	if( state->input_stream->audio_num_channels != 1 ) 
	{
		while (to_count > 0) 
		{
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			to_count -= 2;
		}
	}
	else
	{
		while (to_count > 0) 
		{
			__put_user( ((char)( (*dma_ptr++) >> shift )) ^ 0x80, user_ptr++ );
			dma_ptr++;	/* skip right channel sample */
			to_count--;
		}
	}
	
	return total_to_count;
}

/*
 * Returns negative for error
 * Returns # of bytes transferred out of the from buffer
 * for success.
 */
static  int copy_to_user_with_conversion
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount
)
{
	int ret=0;
	
	if( toCount == 0 )
	{
		printk("copy_to_user_with_conversion - nothing to copy!\n");
	}

	/*
	 * Compact Mode means that each 32 bit word has both the
	 * left and right sample in it.
	 */
	DPRINTK("copy %x %x\n",state->bCompactMode,state->input_stream->audio_format);
	if( state->bCompactMode )
	{
		switch( state->input_stream->audio_format )
		{
			case /*AFMT_U8:*/SNDRV_PCM_FORMAT_U8:
				ret = copy_to_user_U8_CM( state, to, from, toCount );
				break;

			case /*AFMT_S16_LE:*/SNDRV_PCM_FORMAT_S16_LE:
				ret = copy_to_user_S16_LE_CM( state, to, from, toCount );
				break;
		
			case /*AFMT_S8:*/SNDRV_PCM_FORMAT_S8:
				ret = copy_to_user_S8_CM( state, to, from, toCount );
				break;
				
			case /*AFMT_U16_LE:*/SNDRV_PCM_FORMAT_U16_LE:
				ret = copy_to_user_U16_LE_CM( state, to, from, toCount );
				break;
				
			case /*AFMT_S32_BLOCKED:*/SNDRV_PCM_FORMAT_S32_LE:
			default:
				break;
		}
	}
	else
	{
		switch( state->input_stream->audio_format )
		{
			case /*AFMT_U8:*/SNDRV_PCM_FORMAT_U8:
				ret = copy_to_user_U8( state, to, from, toCount );
				break;

			case /*AFMT_S16_LE:*/SNDRV_PCM_FORMAT_S16_LE:
				ret = copy_to_user_S16_LE( state, to, from, toCount );
				break;
		
			case /*AFMT_S8:*/SNDRV_PCM_FORMAT_S8:
				ret = copy_to_user_S8( state, to, from, toCount );
				break;
				
			case /*AFMT_U16_LE:*/SNDRV_PCM_FORMAT_U16_LE:
				ret = copy_to_user_U16_LE( state, to, from, toCount );
				break;
				
			case /*AFMT_S32_BLOCKED:*/SNDRV_PCM_FORMAT_S32_LE:
			default:
				__copy_to_user( (char *)to, from, toCount);
				ret = toCount;
				break;
		}
	}

	return ret;
}


static int copy_from_user_U16_LE_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	unsigned short data;
	unsigned int val;
	int toCount0 = toCount;
	unsigned int * p = (unsigned int *)to;
	unsigned short * userPtr = (unsigned short *)from;
	int mono = (state->output_stream->audio_num_channels == 1);

	/*
	 * Compact mode - slot 3 (left) is the lower 16 bits.  slot 4 (right)
	 * is the upper 16 bits.
	 */
	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = (unsigned int)data & 0x0000ffff;
		
		/*
		 * If mono, use this mono sample for both channels, else we just got
		 * the left channel sample.  So get the right sample and use it for 
		 * the right channel.
		 */
		if( !mono )
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
		}

		*p++ = (((unsigned int)data << 16) | val) ^ 0x80008000;

		toCount -= 4;
	}

	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in one stereo 32 bit sample going out.  So
	 * mono   = (2 bytes in)/(4 bytes out)
	 * stereo = (4 bytes in)/(4 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 2;
	}
	
	return toCount0;
}

static int copy_from_user_U16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	unsigned short data;
	int val;
	int toCount0 = toCount;
	int * p = (int *)to;
	unsigned short * userPtr = (unsigned short *)from;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = state->DAC_bit_width - 16;

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = (data ^ 0x8000) << shift;
		*p++ = val;
		toCount -= 4;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
			val = (data ^ 0x8000) << shift;
		}

		*p++ = val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in a two 32 bit (left & right) samples going out.  So
	 * mono   = (2 bytes in)/(8 bytes out)
	 * stereo = (4 bytes in)/(8 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 4;
	}
	
	return toCount0 / 2;
}

static int copy_from_user_S16_LE_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	short data;
	unsigned int val;
	unsigned int * p = (int *)to;
	int toCount0 = toCount;
	unsigned short * userPtr = (unsigned short *)from;
	int mono = (state->output_stream->audio_num_channels == 1);

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = (unsigned int)data & 0x0000ffff;

		if( !mono )
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
		}

		*p++ = ((unsigned int)data << 16) | val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in one stereo 32 bit sample going out.  So
	 * mono   = (2 bytes in)/(4 bytes out)
	 * stereo = (4 bytes in)/(4 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 2;
	}
	
	return toCount0;
}


static int copy_from_user_S16_LE
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	short data;
	int val;
	int toCount0 = toCount;
	int * p = (int *)to;
	unsigned short * userPtr = (unsigned short *)from;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = state->DAC_bit_width - 16;

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = data << shift;
		*p++ = val;
		toCount -= 4;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
			val = data << shift;
		}

		*p++ = val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 16 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (2 bytes in)/(8 bytes out)
	 * stereo = (4 bytes in)/(8 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 4;
	}
	
	return toCount0 / 2;
}

static int copy_from_user_S8_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	char data;
	int val;
	int toCount0 = toCount;
	int * p = (int *)to;
	unsigned char * userPtr = (unsigned char *)from;
	int mono = (state->output_stream->audio_num_channels == 1);

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = (data << 8) & 0x0000ffff;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
		}

		*p++ = (data << 24) | val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in one stereo 32 bit sample going out.  So
	 * mono   = (1 byte  in)/(4 bytes out)
	 * stereo = (2 bytes in)/(4 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 4;
	}
	
	return toCount0 / 2;
}

static int copy_from_user_S8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	char data;
	int val;
	int toCount0 = toCount;
	int * p = (int *)to;
	unsigned char * userPtr = (unsigned char *)from;
	int mono = (state->output_stream->audio_num_channels == 1);
	int shift = state->DAC_bit_width - 8;

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = data << shift;
		*p++ = val;
		toCount -= 4;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
			val = data << shift;
		}

		*p++ = val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (1 byte  in)/(8 bytes out)
	 * stereo = (2 bytes in)/(8 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 8;
	}
	
	return toCount0 / 4;
}

static int copy_from_user_U8_CM
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int * p = (int *)to;
	unsigned char * userPtr = (unsigned char *)from;
	unsigned char data;
	int val;
	int toCount0 = toCount;

	int mono = (state->output_stream->audio_num_channels == 1);

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = ((data ^ 0x80) << 8) & 0x0000ffff;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
		}

		*p++ = ((data ^ 0x80) << 24) | val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in one stereo 32 bit sample going out.  So
	 * mono   = (1 byte  in)/(4 bytes out)
	 * stereo = (2 bytes in)/(4 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 4;
	}
	
	return toCount0 / 2;
}

static int copy_from_user_U8
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int * p = (int *)to;
	unsigned char * userPtr = (unsigned char *)from;

	unsigned char data;
	int val;
	int toCount0 = toCount;
	int shift = state->DAC_bit_width - 8;

	int mono = (state->output_stream->audio_num_channels == 1);

	while (toCount > 0) 
	{
		if (__get_user(data, userPtr++))
		{
			return -EFAULT;
		}
		val = (data ^ 0x80) << shift;
		*p++ = val;
		toCount -= 4;

		if (!mono) 
		{
			if (__get_user(data, userPtr++))
			{
				return -EFAULT;
			}
			val = (data ^ 0x80) << shift;
		}

		*p++ = val;
		toCount -= 4;
	}

	/*
	 * Each mono or stereo 8 bit sample going in results
	 * in a two 32 bit samples (left & right) going out.  So
	 * mono   = (1 byte  in)/(8 bytes out)
	 * stereo = (2 bytes in)/(8 bytes out)
	 */
	if( mono )
	{
		return toCount0 / 8;
	}
	
	return toCount0 / 4;
}

static int copy_from_user_U32
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	if (copy_from_user( (char *)to, from, toCount)) 
	{
		return -EFAULT;
	}

	return toCount;
}

/*
 * Returns negative for error
 * Returns # of bytes transferred out of the from buffer
 * for success.
 */
static int copy_n_convert
(
	audio_state_t *state,
	const char *to, 
	const char *from, 
	int toCount 
)
{
	int ret=0;
	
	if( toCount == 0 )
	{
		printk("copy_n_convert - nothing to copy!\n");
	}

	/*
	 * Compact Mode means that each 32 bit word has both the
	 * left and right sample in it.
	 */
	if( state->bCompactMode )
	{
		switch( state->output_stream->audio_format )
		{
			case /*AFMT_U8*/ SNDRV_PCM_FORMAT_U8:
				ret = copy_from_user_U8_CM( state, to, from, toCount );
				break;

			case /*AFMT_S16_LE*/ SNDRV_PCM_FORMAT_S16_LE:
				ret = copy_from_user_S16_LE_CM( state, to, from, toCount );
				break;
		
			case /*AFMT_S8*/ SNDRV_PCM_FORMAT_S8:
				ret = copy_from_user_S8_CM( state, to, from, toCount );
				break;
				
			case /*AFMT_U16_LE*/ SNDRV_PCM_FORMAT_U16_LE:
				ret = copy_from_user_U16_LE_CM( state, to, from, toCount );
				break;
				
			case /*AFMT_S32_BLOCKED*/SNDRV_PCM_FORMAT_S32_LE:
			default:
				break;
		}
	}
	else
	{
		switch( state->output_stream->audio_format )
		{
			case /*AFMT_U8*/ SNDRV_PCM_FORMAT_U8:
				ret = copy_from_user_U8( state, to, from, toCount );
				break;

			case /*AFMT_S16_LE*/ SNDRV_PCM_FORMAT_S16_LE:
				ret = copy_from_user_S16_LE( state, to, from, toCount );
				break;
		
			case /*AFMT_S8*/ SNDRV_PCM_FORMAT_S8:
				ret = copy_from_user_S8( state, to, from, toCount );
				break;
				
			case /*AFMT_U16_LE*/ SNDRV_PCM_FORMAT_U16_LE:
				ret = copy_from_user_U16_LE( state, to, from, toCount );
				break;
				
			case /*AFMT_S32_BLOCKED*/SNDRV_PCM_FORMAT_S32_LE:
			default:
				ret = copy_from_user_U32( state, to, from, toCount );
				break;
		}
	}
	
	if( ret == -EFAULT )
	{
		printk( "copy_n_convert: get_user failed!!!!\n" );
	}
	return ret;
}


/*
 * We samples of arbitrary format to be 32 bit for our hardware.
 *	Input format is arbitrary.
 *	Output format is either CM (compact mode) where two 16 bit samples
 *  together in a 32 bit word become a stereo sample or non-CM where
 *  each channel gets a 32 bit word.
 *
 *  Input format       Input sample     Output sample size    ratio (out:in)
 *  bits   channels    size (bytes)       CM   non-CM          CM   non-CM
 *   8       mono          1               4      8            4:1   8:1
 *   8      stereo         2			   4      8            2:1   4:1
 *   16      mono          2			   4      8            2:1   4:1
 *   16     stereo         4			   4      8            1:1   2:1
 *
 *   24      mono          3			   4      8             X    8:3
 *   24     stereo         6			   4      8             X    8:6
 *   32      mono          4			   4      8             X    2:1
 *   32     stereo         8			   4      8             X    1:1
 */
static int scale_to_expanded_samples( audio_state_t *state, int input_bufsize,int dir)
{
	int output_bufsize = input_bufsize;

#ifdef I2S_SCALE_BUFFERS
	audio_stream_t *os ;//= state->output_stream;
	int output_sample_size, input_sample_size;

	if(dir==SNDRV_PCM_STREAM_PLAYBACK)
		os = state->output_stream;
	else
		os = state->input_stream;
	
	if( state->bCompactMode )
	{
		output_sample_size = 4;
	}
	else
	{
		output_sample_size = 8;
	}
	
	input_sample_size = os->audio_stream_bitwidth / 8;
	
	if( os->audio_num_channels != 1 )
	{
		input_sample_size <<= 1;
	}

	output_bufsize = input_bufsize * output_sample_size / input_sample_size;

#endif
	
	return output_bufsize;
}



/*
static int input_scale_to_expanded_samples( audio_state_t *state, int input_bufsize )
{
	int output_bufsize = input_bufsize;

#ifdef I2S_SCALE_BUFFERS
	audio_stream_t *os = state->output_stream;
	int output_sample_size, input_sample_size;
	
	if( state->bCompactMode )
	{
		output_sample_size = 4;
	}
	else
	{
		output_sample_size = 8;
	}
	
	input_sample_size = os->audio_stream_bitwidth / 8;
	
	if( os->audio_num_channels != 1 )
	{
		input_sample_size <<= 1;
	}

	output_bufsize = input_bufsize * input_sample_size /output_sample_size;

#endif
	
	return output_bufsize;
}
*/


/*---------------------------------------------------------------------------------------------*/

/*
 * audio_dma_start
 *
 * Our Ac97 has a specific start order that it likes.  Enable the 
 * Ac97 channel AFTER enabling DMA.  Our I2S is not so picky.
 */
static void audio_dma_start( audio_state_t * state, audio_stream_t * stream )
{
	int err;
	stoptmp =0;
	err = ep93xx_dma_start( stream->dmahandles[0],
					  stream->NumDmaChannels, 
					  stream->dmahandles );
	DPRINTK("audio_dma_start s1=%x\n",err);
	if (state->hw_enable){
		state->hw_enable( state->output_stream == stream );
	}
}

/*
 * audio_dma_pause
 */
static void audio_dma_pause( audio_state_t * state, audio_stream_t * stream )
{
	DPRINTK("audio_dma_pause\n");
	if (state->hw_disable){
		state->hw_disable( state->output_stream == stream );
	}
	
	ep93xx_dma_pause( stream->dmahandles[0],
					  stream->NumDmaChannels, 
					  stream->dmahandles );
}

static void audio_prime_dma( unsigned long externedtotsize,unsigned long totsize,audio_state_t * state, audio_stream_t * s )
{
	unsigned long i=0,j=0,bufferlen=0,bufferlen1=0;
	int ret;

	j=totsize;
	i=externedtotsize;
	//i=s->fragsize*s->nbfrags;
	//i=externedtotsize;
	while(i > 0){
		
//	printk("i=%x\n",i);
	audio_buf_t *buf =s->buf;
	
    //if(i>=s->fragsize)
    //	bufferlen=s->fragsize;
	//else
    //    bufferlen=i;

	if(i>=s->externedfragsize){
	   	bufferlen=s->externedfragsize;
		bufferlen1=s->fragsize;
	}
	else{
	    bufferlen=i;
		bufferlen1=j;
	}
	
//	printk("len=%x\n",bufferlen);
    
	down(&buf->sem);

	ret=ep93xx_dma_add_buffer( s->dmahandles[0], 				/* dma instance*/ 
                               (unsigned int)buf->dma_addr, 				/*source*/
                               0,               							/* dest*/
                               bufferlen,       							/* size*/  
                               0,     										/*is the last chunk?*/   
                               (unsigned int) buf ); 						/* buf id*/
	buf->size=bufferlen;
	buf->reportedsize=bufferlen1;
	memset(buf->start, 0, buf->size);
//	printk("%x ",buf->size);
	buf->sent = 1;
    NEXT_BUF(s, buf);

	up(&buf->sem);
    i=i-bufferlen;
	j=j-bufferlen1;
//  printk("add dma buffer ret=%x,buffer=%x\n",ret,bufferlen);
//	printk("a1\n");  
    }

//      if( !s->active && !s->stopped )
//        {
//                s->active = 1;
//                audio_dma_start( state, s );
//        }





}

/*
 * audio_deallocate_buffers
 * 
 * This function frees all buffers
 */
static void audio_deallocate_buffers( snd_pcm_substream_t *substream, audio_stream_t * s )
{
	int frag, i;
	audio_state_t  *state = (audio_state_t *)substream->private_data;
	
	DPRINTK("EP93xx - audio_deallocate_buffers\n");

	/* ensure DMA won't run anymore */
	audio_dma_pause( state, s );
	s->active = 0;
	s->stopped = 0;

	for( i=0 ; i<s->NumDmaChannels ; i++ )
	{
		ep93xx_dma_flush( s->dmahandles[0] );
	}
	
	if (s->buffers) 
	{
		for (frag = 0; frag < s->nbfrags; frag++) 
		{
			if (!s->buffers[frag].master)
			{
				continue;
			}

			/*
			consistent_free(s->buffers[frag].start,
					s->buffers[frag].master,
					s->buffers[frag].dma_addr);
			*/
			s->buffers[frag].start		=	NULL;
			s->buffers[frag].master		=	0;
			s->buffers[frag].dma_addr	=	(dma_addr_t)NULL;
			
		}

		/*
		 * Free the space allocated to the array of buf structs.
		 */
		kfree(s->buffers);
		s->buffers = NULL;
	}

	s->buf_idx = 0;
	s->buf = NULL;

	DPRINTK("EP93xx - audio_deallocate_buffers - EXIT\n");
}


/*
 * audio_allocate_buffers
 *
 * This function allocates the buffer structure array and buffer data space
 * according to the current number of fragments and fragment size.
 * Note that the output_stream and input_stream structs are allocated
 * in ep93xx-ac97.c or ep93xx-i2s.c.
 */
static int audio_allocate_buffers( snd_pcm_substream_t *substream, audio_stream_t * s)
{
	int frag;
	int dmasize = 0;
	char *dmabuf = NULL;
	dma_addr_t dmaphys = 0;
	
	/*audio_state_t  *state = (audio_state_t *)substream->private_data;*/
	struct snd_dma_buffer *pcmbuf = &substream->dma_buffer;
	
	if (s->buffers){
		printk("%s BUSY\n",__FUNCTION__);
		return -EBUSY;
	}

//	printk("EP93xx audio_allocate_buffers ENTER\n" );
	
	/*
	 * Allocate space for the array of audio_buf_t structs.
	 */
	s->buffers = (audio_buf_t *)
		kmalloc(sizeof(audio_buf_t) * s->nbfrags, GFP_KERNEL);

	if (!s->buffers)
	{
		printk(AUDIO_NAME ": unable to allocate audio memory\n ");
		audio_deallocate_buffers(substream,s);
		return -ENOMEM;
	}

	memset( s->buffers, 0, sizeof(audio_buf_t) * s->nbfrags );

	/*
	 * Let's allocate non-cached memory for DMA buffers.
	 * We try to allocate all memory at once.
	 * If this fails (a common reason is memory fragmentation),
	 * then we allocate more smaller buffers.
	 */
	
	
	for (frag = 0; frag < s->nbfrags; frag++) 
	{
		audio_buf_t *buf = &s->buffers[frag];

		if (!dmasize) 
		{
			/*
			 * First try to allocate enough for all the frags that
			 * don't yet have memory allocated.
			 */
			/*dmasize = (s->nbfrags - frag) * s->fragsize;*/
			//dmasize	=	pcmbuf->bytes; 
			
			if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
				dmasize	=	s->nbfrags*s->externedfragsize;//scale_to_expanded_samples(state,pcmbuf->bytes,substream->stream);
			else
				//dmasize	=	pcmbuf->bytes;
				dmasize	=	s->nbfrags*s->externedfragsize;//scale_to_expanded_samples(state,pcmbuf->bytes,substream->stream);

			
//			printk("pcmbuf->bytes=%x\n",pcmbuf->bytes);
			//dmasize=	substream->runtime->dma_bytes;
			do {
			  	/*dmabuf = consistent_alloc(GFP_KERNEL|GFP_DMA,
							  dmasize, &dmaphys);
				*/
				dmabuf	=	pcmbuf->area;
				//dmabuf	=	substream->runtime->dma_area;

				dmaphys	=	pcmbuf->addr;
				//dmaphys =	substream->runtime->dma_addr;
				/* 
				 * If that fails, try to allocate a chunk of memory
				 * that is one less fragment is size.
				 */
				if (!dmabuf)
				{
					dmasize -= s->fragsize;
				}

			  /* 
			   * Keep trying but the minimum we'll attempt is one 
			   * fragment.  If we can't even get that, give up.
			   */
			} while (!dmabuf && dmasize);
		
			/*
			 * If we do fail to allocate enough for all the frags, 
			 * deallocate whatever we did get and quit.
			 */
			if (!dmabuf)
			{
				DPRINTK(AUDIO_NAME ": unable to allocate audio memory\n ");
				audio_deallocate_buffers(substream, s);
				return -ENOMEM;
			}

//			printk("EP93xx allocated %d bytes:  dmabuf=0x%08x  dmaphys=0x%08x\n",
//					dmasize, (int)dmabuf, (int)dmaphys );
			
			/*
			 * Success!	 Save the size of this chunk to use when we deallocate it.
			 */
			buf->master = dmasize;
			memzero(dmabuf, dmasize);
		}

		/*
		 * Save the start address of the buf fragment.
		 * We know the size of the fragment is fragsize.
		 */
		buf->start = dmabuf;
		buf->dma_addr = dmaphys;
//		printk("buf->start=%x,buf->dma_addr=%x\n",buf->start,buf->dma_addr);
		buf->stream = s;
		buf->size = 0;
		buf->reportedsize = 0;
		buf->sent = 0;

		sema_init(&buf->sem, 1);

		/*
		 * Now if we only allocated the minimal one frag of space, the
		 * dmasize will be ==0 after this subtraction so it will allocate more
		 * for the next frag.  Otherwise, the next time(s) thru this for loop
		 * will dole out frag sized pieces of this big master chunk.
		 */
		//dmabuf += s->fragsize;
		//dmaphys += s->fragsize;
		//dmasize -= s->fragsize;

		dmabuf += s->externedfragsize;
		dmaphys += s->externedfragsize;
		dmasize -= s->externedfragsize;
	}

	/*
	 * Initialize the stream.
	 */
	s->buf_idx = 0;				/* Init the current buffer index. 			*/
	s->buf = &s->buffers[0];	/* Point buf to the current buffer struct. 	*/
	s->bytecount = 0;
	s->externedbytecount = 0;
	s->getptrCount = 0;
	s->fragcount = 0;

//	DPRINTK("EP93xx audio_allocate_buffers -- exit SUCCESS\n" );
	return 0;

}


/*
 * audio_reset_buffers
 *
 * This function stops and flushes the dma, gets all buffers back
 * from the DMA driver and resets them ready to be used again.
 */
/*
static void audio_reset_buffers( audio_state_t * state, audio_stream_t * s )
{
	int frag, i;

	audio_dma_pause( state, s );
	s->active = 0;
	s->stopped = 0;

	for( i=0 ; i < s->NumDmaChannels ; i++ )
	{
		ep93xx_dma_flush( s->dmahandles[0] );
	}

	if (s->buffers) 
	{
		for (frag = 0; frag < s->nbfrags; frag++) 
		{
			audio_buf_t *buf = &s->buffers[frag];
			buf->size = 0;
			buf->sent = 0;
			sema_init(&buf->sem, 1);
		}
	}

	s->bytecount = 0;
	s->getptrCount = 0;
	s->fragcount = 0;
}
*/

/*
 * DMA callback functions
 */
static void audio_dma_tx_callback
( 
	ep93xx_dma_int_t DMAInt,
	ep93xx_dma_dev_t device, 
	unsigned int user_data 
)
{
	unsigned int buf_id;
	int handle;


	
	snd_pcm_substream_t *substream	=	(snd_pcm_substream_t *)user_data;
	audio_state_t *state = (audio_state_t *)(substream->private_data);
	audio_stream_t *s = state->output_stream;

//printk("audio_dma_tx_callback--ENTER:device=%x\n",device);	
//printk("tc1\n");
 /*	DPRINTK( "audio_dma_tx_callback  - enter\n"); */
	
	switch( device )
	{
    	case DMATx_I2S3:
			handle = s->dmahandles[2];
			break;

    	case DMATx_I2S2:
			handle = s->dmahandles[1];
			break;

		case DMATx_I2S1:
		default:
			handle = s->dmahandles[0];
			break;
	}
	
	s->mapped=1;


  	if (s->mapped)
	{
		/*
		 * If we are mapped, get one buffer back and recycle it.
		 */
		if( ep93xx_dma_remove_buffer( handle, &buf_id ) >= 0 )
		{
		   	audio_buf_t *buf = (audio_buf_t *)buf_id;
	//		printk("%x\n",buf->dma_addr);
		 /*	DPRINTK( "audio_dma_tx_callback  - got buffer index=%d\n", buf_id); */
		
			DPRINTK("%x\n",*(unsigned int *)buf->start);
			/* Accounting */
			s->bytecount += buf->size;
			s->externedbytecount += buf->reportedsize;
			s->fragcount++;
			//buf->size = 0;
		
			/* Recycle buffer */
		 	//buf->size = s->fragsize;
			
			ep93xx_dma_add_buffer( s->dmahandles[0], 	/* dma instance 		*/
								   (unsigned int)buf->dma_addr, /* source 		*/
								   0, 					/* dest 				*/
								   buf->reportedsize,/*s->fragsize,*/			/* size					*/
								   0,					/* is the last chunk? 	*/
								   (unsigned int) buf ); /* buf id			 	*/	
	  		
			buf->sent = 1;
		}
	}
	else
	{
//		printk("audio_dma_tx_callback no been mapped\n");	
		/*
		 * Get all buffers that are free'ed back and clear their semephores.
		 */
		while( ep93xx_dma_remove_buffer( handle, &buf_id ) >= 0 )
		{
		   	audio_buf_t *buf = (audio_buf_t *)buf_id;
			memset(buf->start, 0, buf->size);
//			printk("audio_dma_tx_callback no been mapped %x,buf->size=%x,s->fragcount=%x\n",
//				buf_id,buf->size,s->fragcount);
		 /*	DPRINTK( "audio_dma_tx_callback  - got buffer index=%d\n", buf_id); */
		
			/* Accounting */
//			printk("%x", buf->size);
			s->bytecount += buf->size;
			s->fragcount++;
		//	buf->size = 0;
			buf->sent = 0;
			
			/*
			 * Release the semaphore on this buf.
			 * If write is waiting on this buf then it can go ahead and fill
			 * it and send it to the dma.
			 */
			up(&buf->sem);
		}

//		printk("audio_dma_tx_callback no been mapped --over\n");
	}

	/* And any process polling on write. */
	wake_up(&s->wq);

	/*update the substream runtime date status*/
	snd_pcm_period_elapsed(substream);

	
//	printk("audio_dma_tx_callback EXIT\n");

	
}

static void audio_dma_rx_callback
( 
	ep93xx_dma_int_t DMAInt,
	ep93xx_dma_dev_t device, 
	unsigned int user_data 
)
{
	unsigned int buf_id;
	int handle;


	snd_pcm_substream_t *substream	=	(snd_pcm_substream_t *)user_data;
	audio_state_t *state = (audio_state_t *)(substream->private_data);
	audio_stream_t *s = state->input_stream;
	audio_buf_t *buf;

//printk("audio_dma_rx_callback\n");
	switch( device )
	{
    	case DMARx_I2S3:
			handle = s->dmahandles[2];
			break;

    	case DMARx_I2S2:
			handle = s->dmahandles[1];
			break;

		case DMARx_I2S1:
		default:
			handle = s->dmahandles[0];
			break;
	}

	//s->mapped=1;
	/*
	 * Keep removing and recycling buffers as long as there are buffers
	 * to remove.
	 */
	while( !ep93xx_dma_remove_buffer( handle, &buf_id ) )
	{
	   	buf = (audio_buf_t *) buf_id;
		DPRINTK("ReadEnd %x\n",ReadEnd);
		Stack_buf_id[ReadEnd]	=	buf_id;
		ReadEnd	=	(ReadEnd	+1)%(s->nbfrags);
		s->buf = (audio_buf_t *) buf_id;
//		printk("%x\n",*(unsigned int *)buf->start);
   
		//while(stoptmp==1){}


		/* Accounting */
//		s->bytecount += buf->size;
		s->fragcount++;
//		s->buf->size=buf->size;
//		printk("bufsize=%x %x\n",s->buf->size,buf->size);
		/* Recycle buffer */
		if( s->mapped ) 
		{
//			if(s->stopped==0){	
            		down(&buf->sem);
		  	//s->bytecount += buf->size;
			s->bytecount += buf->reportedsize;
			s->externedbytecount += buf->size;

		DPRINTK("rx1\n");	
			snd_pcm_period_elapsed(substream);
		DPRINTK("rx2\n");	

		 	ep93xx_dma_add_buffer( s->dmahandles[0], 	/* dma instance 		*/
								   (unsigned int)buf->dma_addr, /* source 		*/
								   0, 					/* dest 				*/
								   buf->size,	/*s->fragsize,*/	/* size 	*/
								   0,					/* is the last chunk? 	*/
								   (unsigned int) buf ); /* buf id			 	*/	

			up(&buf->sem);
//			}

 		} 
		else 
		{
//			snd_pcm_period_elapsed(substream);

	//		buf->size = s->fragsize;
	//		printk("up\n");
//			up(&s->buf->sem);
			
		//	while(stoptmp==1){}			

			down(&s->buf->sem);
	
			//s->bytecount += buf->size;
			s->bytecount += buf->reportedsize;
			s->externedbytecount += buf->size;

			up(&s->buf->sem);
		}

		/* And any process polling on write. */
		wake_up(&s->wq);
	}

	/*shrek add it*/
	//down(&sLoopsem);
	DPRINTK("update 1\n");
	snd_pcm_period_elapsed(substream);
	DPRINTK("update 2\n");
	//memset(buf->start, 0, buf->size);
}


/*
 * audio_sync
 *
 * Wait until the last byte written to this device has been played.
 */
static int audio_sync(snd_pcm_substream_t *substream)
{
	audio_state_t *state = (audio_state_t *)substream->private_data;
	audio_stream_t *s = state->output_stream;
	audio_buf_t *buf;
/*	int frag; */
	int buf_wait_index;
  /*	audio_buf_t *buftemp; */

	DPRINTK( "EP93xx - audio_sync - enter\n" );

	if (!(substream->stream == SNDRV_PCM_STREAM_PLAYBACK ) || !s->buffers || s->mapped)
	{
		printk( "EP93xx - audio_sync - exit immediately.\n" );
		return 0;
	}
	
	/*
	 * Send current buffer if it contains data and hasn't been sent. 
	 */
#if 0	 
	for (frag = 0; frag < s->nbfrags; frag++) 
	{
		buftemp = &s->buffers[frag];
		DPRINTK("audio_sync - buf# %d: size=%d  sent=%d\n",
			frag, buftemp->size, buftemp->sent );
	}
#endif

	buf = s->buf;
	if( buf->size && !buf->sent ) 
	{
		DPRINTK("audio_sync -- SENDING BUFFER index=%d size=%d\n",
				s->buf_idx, buf->size );
		
		down(&buf->sem);
	 	ep93xx_dma_add_buffer( s->dmahandles[0], 	/* dma instance 		*/
							   (unsigned int)buf->dma_addr, /* source 		*/
							   0, 					/* dest 				*/
							   buf->size,			/* size					*/
							   0,					/* is the last chunk? 	*/
							   (unsigned int) buf ); /* buf id			 	*/	
	 	buf->sent = 1;
		NEXT_BUF(s, buf);
	}

	/*
	 * Let's wait for the last buffer we sent i.e. the one before the
	 * current buf_idx.  When we acquire the semaphore, this means either:
	 * - DMA on the buffer completed or
	 * - the buffer was already free thus nothing else to sync.
	 */
	buf_wait_index = ((s->nbfrags + s->buf_idx - 1) % s->nbfrags);
	buf = s->buffers + buf_wait_index;

#if 0
	while( buf->sent )
	{
		mdelay( 1000 );
		for (frag = 0; frag < s->nbfrags; frag++) 
		{
			buftemp = &s->buffers[frag];
			DPRINTK("audio_sync - buf# %d: size=%d  sent=%d\n",
				frag, buftemp->size, buftemp->sent );
		}
		DPRINTK("\n");
	}
	DPRINTK("audio_sync -- current buf index is %d.  Waiting on buffer index=%d\n",
			s->buf_idx, buf_wait_index );
#endif

	if (down_interruptible(&buf->sem))
	{
		return -EINTR;
	}
	
	up(&buf->sem);
	DPRINTK("EP93xx - audio_sync - EXIT\n");
	return 0;
}



/*
*
*/
static int audio_release(snd_pcm_substream_t *substream)
{
	int i;
	audio_state_t *state = (audio_state_t *)substream->private_data;

	DPRINTK("EP93xx - audio_release -- enter.  \n");
			

	down(&state->sem);

	/*read data from codec to controller */
	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE) 
	{
		audio_deallocate_buffers(substream, state->input_stream);
		
		for( i=0 ; i < state->input_stream->NumDmaChannels ; i++ )
		{
			ep93xx_dma_free( state->input_stream->dmahandles[i] );
		}
		state->rd_ref = 0;
	}
	
	/*write data from controller to codec*/
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK ) 
	{
		/*
		 * Wait for all our buffers to play.
		 */
		audio_sync(substream);
		
		/*
		 * audio_deallocate_buffers will pause and flush the dma
		 * then deallocate the buffers.
		 */
		audio_deallocate_buffers(substream, state->output_stream);
		for( i=0 ; i < state->output_stream->NumDmaChannels ; i++ )
		{
			ep93xx_dma_free( state->output_stream->dmahandles[i] );
		}
		state->wr_ref = 0;
	}

	up(&state->sem);

	DPRINTK("EP93xx - audio_release -- EXIT\n");

	return 0;
}

/*--------------------------------------------------------------------------------------------*/

/*
 *ep93xx_ac97_pcm_hw_params
 *set the pcm dma buffer to the runtime pcm dma buffer
 */
static int ep93xx_ac97_pcm_hw_params(snd_pcm_substream_t *substream,
				snd_pcm_hw_params_t *params)
{
	snd_pcm_runtime_t 	*runtime 		= substream->runtime;
	
	audio_state_t  		*state 			= (audio_state_t *)substream->private_data;
	
	audio_stream_t 		*sTriggerStream = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
											state->output_stream:state->input_stream;

	/*
	size_t totsize = params_buffer_bytes(params);
	size_t period  = params_period_bytes(params);
	struct snd_dma_buffer *pcmbuf = &substream->dma_buffer;
	*/
	int err;
	
	DPRINTK("ep93xx_ac97_pcm_hw_params----ENTER\n");
	DPRINTK("%s: sub=%p, hwp=%p\n", __FUNCTION__, substream, params);
	DPRINTK("%s:PCM addr=%x, area=%x,size=%x\n", __FUNCTION__, pcmbuf->addr, pcmbuf->area,pcmbuf->bytes);
	/*
	 *Set the substream DMA buffer address to the runtime DMA buffer address
	 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	snd_pcm_set_runtime_buffer(substream, &substream->dma_buffer);
	return 0;
#else
	err = snd_pcm_lib_malloc_pages(substream,params_buffer_bytes(params));
	if(err<0){
		printk("err=%x,size=%x\n",err,params_buffer_bytes(params));
	}
	
	DPRINTK("err=%x,size=%x\n",err,params_buffer_bytes(params));

	DPRINTK("%s:Runtime : addr=%x, area=%x,size=%x\n", __FUNCTION__, 
		runtime->dma_addr, runtime->dma_area,runtime->dma_bytes);

	DPRINTK("%s:Runtime frame_bits=%x, buffer_size=%x,periods=%x,period_size=%x,period_step=%x,rate=%x,format=%x\n",
		__FUNCTION__, 
		runtime->frame_bits, runtime->buffer_size,runtime->periods,runtime->period_size,
		runtime->period_step,runtime->rate,runtime->format);
	DPRINTK("\ntotsize=%x,period=%x\n",totsize,period);
	
	sTriggerStream->fragsize=runtime->period_size;
	sTriggerStream->nbfrags	=runtime->periods;
	
	return err;
#endif

	
}

/*
 *free the runtime_data buffer 
 */
static int ep93xx_ac97_pcm_hw_free(snd_pcm_substream_t *substream)
{

	DPRINTK("ep93xx_ac97_pcm_hw_free----ENTER\n");
	/*
	struct pxa2xx_runtime_data *rtd = substream->runtime->private_data;

	*rtd->params->drcmr = 0;
	*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	snd_pcm_set_runtime_buffer(substream, NULL);
	return 0;
#else
	return snd_pcm_lib_free_pages(substream);
#endif

}


/*
 *ep93xx_ac97_pcm_prepare: need to finish these functions as lower
 *chip_set_sample_format
 *chip_set_sample_rate
 *chip_set_channels
 *chip_set_dma_setup
 */
static int ep93xx_ac97_pcm_prepare(snd_pcm_substream_t *substream)
{
	audio_state_t *state  		= (audio_state_t *)substream->private_data;

	snd_pcm_runtime_t *runtime 	= substream->runtime;
	
	audio_stream_t *sTriggerStream = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
					state->output_stream:state->input_stream;
	int ret	=	0;

	size_t totsize;//= snd_pcm_lib_buffer_bytes(substream);
	size_t externedtotsize;//	= scale_to_expanded_samples(state,snd_pcm_lib_buffer_bytes(substream));




	
	DPRINTK("ep93xx_ac97_pcm_prepare----ENTER\n");

	DPRINTK("%s:Runtime channel=%x,frame_bits=%x, buffer_size=%x,periods=%x,period_size=%x,period_step=%x,rate=%x\n,format=%x,snd_pcm_lib_period_bytes(substream)=%x\n",
	    __FUNCTION__,runtime->channels,
	    runtime->frame_bits, runtime->buffer_size,runtime->periods,runtime->period_size,
	    runtime->period_step,runtime->rate,runtime->format,snd_pcm_lib_period_bytes(substream));

	DPRINTK("%s:Runtime addr=%x, area=%x,size=%x\n", __FUNCTION__, 
		runtime->dma_addr, runtime->dma_area,runtime->dma_bytes);
	DPRINTK("%s:PCM addr=%x, area=%x,size=%x\n", __FUNCTION__, 
		substream->dma_buffer.addr,substream->dma_buffer.area,substream->dma_buffer.bytes);


	/*
	 *Re_init the pcm frame_bits,rate,format,channels for AC97 Controller and Codec
	 */

	/*
	 *Do the set channel for audio streaming.
	 */
	sTriggerStream->audio_num_channels	=	runtime->channels;
	/*
	 * Do the set format for audio streaming.
	 */
	audio_set_format( sTriggerStream, (runtime->format));
	 
	/*
	 * If the audio hardware needs some setup, do it.
	 */
	if( state->set_hw_serial_format )
	{
		state->set_hw_serial_format( sTriggerStream->audio_format );
	}
	
	ret	=	state->prepare(substream);


	totsize		= snd_pcm_lib_buffer_bytes(substream);
	externedtotsize	= scale_to_expanded_samples(state,snd_pcm_lib_buffer_bytes(substream),substream->stream);
	
	sTriggerStream->fragsize		=	snd_pcm_lib_period_bytes(substream);
	sTriggerStream->externedfragsize	=	scale_to_expanded_samples(state,snd_pcm_lib_period_bytes(substream),substream->stream );
	
	if(totsize%sTriggerStream->fragsize==0)
		sTriggerStream->nbfrags = totsize/sTriggerStream->fragsize;
	else
		sTriggerStream->nbfrags = (totsize/sTriggerStream->fragsize) +1;


	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE){
		DPRINTK("EP93xx - audio_read Enter\n");
		if (sTriggerStream->mapped){
			printk("mmap error\n");
			return -ENXIO;
		}
		if (!sTriggerStream->active) 
		{
			if (!sTriggerStream->buffers && audio_allocate_buffers(substream, sTriggerStream))
			{
				printk("memory error\n");
				return -ENOMEM;
			}
			audio_prime_dma( externedtotsize,totsize,state, sTriggerStream);
		}

		DPRINTK("EP93xx - audio_read Exit\n");
	}

	return ret;
}


/*
 *start/stop/pause dma translate
 */
static int ep93xx_ac97_pcm_trigger(snd_pcm_substream_t *substream, int cmd)
{
	
	audio_state_t  *state = (audio_state_t *)substream->private_data;
	audio_stream_t *sTriggerStream = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
					state->output_stream:state->input_stream;
	
	size_t totsize= snd_pcm_lib_buffer_bytes(substream),bufferlen=0,bufferlen1=0;
	size_t externedtotsize	= scale_to_expanded_samples(state,snd_pcm_lib_buffer_bytes(substream),substream->stream);

	//snd_pcm_runtime_t *runtime = substream->runtime;	
	int ret = 0,i=0,j=0;
	unsigned long flags;

	//unsigned int bufdma_addr=substream->dma_buffer.addr;
	//unsigned int bufdma_addr=substream->runtime->dma_addr;
 /* 
 	printk("%s:Runtime sample_bits=%x,byte_align=%x,frame_bits=%x,\nbuffer_size=%x,periods=%x,period_size=%x,period_step=%x,rate=%x\n,format=%x,snd_pcm_lib_period_bytes(substream)=%x,totsize=%x\n",
            __FUNCTION__,runtime->sample_bits,runtime->byte_align,
            runtime->frame_bits, runtime->buffer_size,runtime->periods,runtime->period_size,
            runtime->period_step,runtime->rate,runtime->format,snd_pcm_lib_period_bytes(substream),totsize);
  */
  
	/*
	 *Init the substream privat data struct audio_stream_t
	 *runtime->period_size based on frame not byes
	 */
	sTriggerStream->fragsize		=	snd_pcm_lib_period_bytes(substream);
	sTriggerStream->externedfragsize=	scale_to_expanded_samples(state,snd_pcm_lib_period_bytes(substream),substream->stream);
	
	if(totsize%sTriggerStream->fragsize==0)
		sTriggerStream->nbfrags = totsize/sTriggerStream->fragsize;
	else
		sTriggerStream->nbfrags = (totsize/sTriggerStream->fragsize) +1;
	
//printk("fragsize=%x,externedfragsize=%x,totsize=%x,externedtotsize=%x\n",
//		sTriggerStream->fragsize,sTriggerStream->externedfragsize,
//		totsize,externedtotsize);
/*	
printk("ep93xx_ac97_pcm_trigger----ENTER cmd=%x,active=%x,mapped=%x,stopped=%x\n",
		cmd,sTriggerStream->active,sTriggerStream->mapped,sTriggerStream->stopped);

printk("addr=%x,area=%x,bytes=%x\n",
		substream->dma_buffer.addr,substream->dma_buffer.area,
		substream->dma_buffer.bytes);
printk("data=%x,%x,%x,%x,%x,%x,%x,%x\n",
		*(substream->dma_buffer.area),*(substream->dma_buffer.area+1),
		*(substream->dma_buffer.area+2),*(substream->dma_buffer.area+3),
		*(substream->dma_buffer.area+4),*(substream->dma_buffer.area+5),
		*(substream->dma_buffer.area+6),*(substream->dma_buffer.area+7));
*/
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK){
		if (sTriggerStream->mapped) {
			printk("%s error %i\n", __FUNCTION__, __LINE__);
			return -ENXIO;
		}

		/*
		 * Allocate dma buffers if we haven't already done so.
		 */
		if( !sTriggerStream->buffers && audio_allocate_buffers(substream, sTriggerStream) )
		{
			printk("%s error %i\n", __FUNCTION__, __LINE__);
			return -ENOMEM;
		}
		
		sTriggerStream->buf_idx=0;
		/*
		 * If we haven't already started the DMA start it.
		 * But don't start it if we are waiting on a trigger.
		 */
		if( !sTriggerStream->active && !sTriggerStream->stopped )
		{
	//	 	printk("ep93xx_dma_start \n");

			sTriggerStream->active = 1;
			audio_dma_start( state, sTriggerStream );
		}

		/*
		 *Add the audio data to dma request
		 */
		local_irq_save(flags);
		
		//i=totsize;
		j=totsize;
		i=externedtotsize;
		while(i > 0){
			
		DPRINTK("i=%x\n",i);
		audio_buf_t *buf =sTriggerStream->buf;
		
	    //if(i>=sTriggerStream->fragsize)
	    //	bufferlen=sTriggerStream->fragsize;
		if(i>=sTriggerStream->externedfragsize){
	    	bufferlen1=sTriggerStream->externedfragsize;
			bufferlen=sTriggerStream->fragsize;
		}
	    else{
	        bufferlen1=i;
			bufferlen=j;
	    }
		
		DPRINTK("len=%x\n",bufferlen);
	    
		down(&buf->sem);

		ret=ep93xx_dma_add_buffer( sTriggerStream->dmahandles[0], 				/* dma instance*/ 
	                               (unsigned int)buf->dma_addr, 				/*source*/
	                               0,               							/* dest*/
	                               bufferlen,       							/* size*/  
	                               0,     										/*is the last chunk?*/   
	                               (unsigned int) buf ); 						/* buf id*/
		buf->size=bufferlen;
		buf->reportedsize=bufferlen1;
			
		DPRINTK("%x ",buf->size);
		buf->sent = 1;
	    NEXT_BUF(sTriggerStream, buf);

	    //i=i-bufferlen;
	   	i=i-bufferlen1;
		j=j-bufferlen;
	//  printk("add dma buffer ret=%x,buffer=%x\n",ret,bufferlen);
		DPRINTK("a1\n");  
	    }

		local_irq_restore(flags);
	//	printk("triger over\n");
	}
	else{
		DPRINTK("triger\n");
		
	    if( !sTriggerStream->active && !sTriggerStream->stopped )
        {
                sTriggerStream->active = 1;
                audio_dma_start( state, sTriggerStream );
        }

/*		if (sTriggerStream->mapped){
			DPRINTK("mmap error\n");
			return -ENXIO;
		}
		if (!sTriggerStream->active) 
		{
			if (!sTriggerStream->buffers && audio_allocate_buffers(substream, sTriggerStream))
			{
				DPRINTK("memory error\n");
				return -ENOMEM;
			}
			audio_prime_dma( totsize,state, sTriggerStream);
		}

		DPRINTK("EP93xx - audio_read Exit\n");
*/
	}
	
	break;

	case SNDRV_PCM_TRIGGER_STOP:
	case SNDRV_PCM_TRIGGER_SUSPEND:
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:

		audio_dma_pause( state, sTriggerStream );
		sTriggerStream->stopped = 1;
		sTriggerStream->active 	= 0;
		break;

	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
/*		if (!sTriggerStream->active){
			if (!sTriggerStream->buffers && audio_allocate_buffers(substream, sTriggerStream))
				return -ENOMEM;
			
			if (sTriggerStream->mapped)
				audio_prime_dma( totsize,state, sTriggerStream);
		}
					
		if (sTriggerStream->stopped){
			sTriggerStream->stopped = 0;
			sTriggerStream->active 	= 1;
			audio_dma_start( state, sTriggerStream );
		}
*/		break;

	default:
		ret = -EINVAL;
		DPRINTK("SNDRV_PCM_TRIGGER didnot support this command %x",cmd);
	}

	return ret;
}

/*
 *Get the frame location
 */
static snd_pcm_uframes_t ep93xx_ac97_pcm_pointer(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime 	= substream->runtime;

	audio_state_t  *state 		= (audio_state_t *)substream->private_data;
	audio_stream_t *sStream 	= (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
									state->output_stream:state->input_stream;
	snd_pcm_uframes_t x=0;

	unsigned int buf_id, total,current_frac;
	/*unsigned int temp=0;*/


	if( (ep93xx_dma_get_position(sStream->dmahandles[0], &buf_id, &total, &current_frac) == 0) &&
		(buf_id != -1) ){
		 //x = bytes_to_frames(runtime, total/*sStream->fragcount*(sStream->fragsize) + current_frac*/);
		 if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
    			x = bytes_to_frames(runtime,sStream->bytecount);
		 else
		 	x = bytes_to_frames(runtime,sStream->bytecount);
		//x = bytes_to_frames(runtime,sStream->externedbytecount); 
	}
	else{
//		x = bytes_to_frames(runtime, total/*sStream->fragcount*(sStream->fragsize)*/ );
 		//x = bytes_to_frames(runtime,sStream->bytecount);
		 if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
	   		x = bytes_to_frames(runtime,sStream->bytecount);
		 else
		 	x = bytes_to_frames(runtime,sStream->bytecount);
		//x = bytes_to_frames(runtime,sStream->externedbytecount);
	}
	
	DPRINTK(" %x %x %x\n",sStream->bytecount,current_frac,bytes_to_frames(runtime,total)%runtime->buffer_size);
	if (x >= runtime->buffer_size){
	//if(x>=bytes_to_frames(runtime,sStream->fragsize)*sStream->nbfrags){
//  		x = 0;
		x=runtime->buffer_size;
	   	sStream->bytecount=0;
		sStream->externedbytecount=0;
    }
	DPRINTK("x=%x\n",x);




	return x;
}



static int ep93xx_ac97_pcm_open(snd_pcm_substream_t *substream)
{
	audio_state_t *state = substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	
	audio_stream_t *os = state->output_stream;
	audio_stream_t *is = state->input_stream;
	int err;

	DPRINTK("ep93xx_ac97_pcm_open----ENTER\n");
	
	/* set the hardware description*/
	runtime->hw = ep93xx_ac97_pcm_hardware;


	down(&state->sem);
	
	/*write :output direction*/
	if(substream->stream == SNDRV_PCM_STREAM_PLAYBACK ) 
	{
		DPRINTK("ep93xx_audio_open -- FMODE_WRITE\n");
		
		strncpy( os->devicename, state->output_id, MAX_DEVICE_NAME );

		/* 
	 	 * Request DMA channels 
	 	 */
		err = ep93xx_dma_request( 	&os->dmahandles[0], 
									os->devicename,
                       				state->output_dma );
		if (err)
		{
			up(&state->sem);
			DPRINTK("ep93xx_audio_open -- EXIT ERROR dma request failed\n");
			return err;
		}
		/* 
	 	 * Config DMA channels 
	 	 */
		err = ep93xx_dma_config( 	os->dmahandles[0], 
									IGNORE_CHANNEL_ERROR, 
									0,
									audio_dma_tx_callback, 
									(unsigned int)substream );
		if (err) 
		{
			ep93xx_dma_free( os->dmahandles[0] );
			up(&state->sem);
			DPRINTK("ep93xx_audio_attach -- EXIT ERROR dma config failed\n");
			return err;
		}

		os->NumDmaChannels = 1;
		state->wr_ref = 1;
		audio_deallocate_buffers(substream,os);

		os->fragsize 	= AUDIO_FRAGSIZE_DEFAULT;
		os->nbfrags  	= AUDIO_NBFRAGS_DEFAULT;

		os->reportedfrags = scale_to_expanded_samples( state, AUDIO_NBFRAGS_DEFAULT,substream->stream );

		os->mapped 	= 0;
		os->stopped = 0;
		init_waitqueue_head(&os->wq);
	}
	
	/*read :input direction*/
	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE) 
	{
		strncpy( is->devicename, state->input_id, MAX_DEVICE_NAME );
		/* 
	 	 * Request DMA channels 
	 	 */
		err = ep93xx_dma_request( 	&is->dmahandles[0], 
									is->devicename,
                       				state->input_dma);
		DPRINTK("%x %x %x\n", &is->dmahandles[0],is->devicename,state->input_dma);
		if (err) 
		{
			up(&state->sem);
			printk("ep93xx_audio_attach -- EXIT ERROR dma request failed\n");
			return err;
		}
		
		/* 
	 	 * Config DMA channels 
	 	 */
		err = ep93xx_dma_config( 	is->dmahandles[0], 
									IGNORE_CHANNEL_ERROR, 
									0,
									audio_dma_rx_callback, 
									(unsigned int)substream );
		if (err) 
		{
			ep93xx_dma_free( is->dmahandles[0] );
			up(&state->sem);
			printk("ep93xx_audio_attach -- EXIT ERROR dma config failed\n");
			return err;
		}

		is->NumDmaChannels = 1;
		state->rd_ref = 1;
		audio_deallocate_buffers(substream,is);
		is->fragsize 	= AUDIO_FRAGSIZE_DEFAULT;
		is->nbfrags 	= AUDIO_NBFRAGS_DEFAULT;

		os->reportedfrags = scale_to_expanded_samples( state, AUDIO_NBFRAGS_DEFAULT,substream->stream );

		is->mapped 	= 0;
		is->stopped = 		0;
		init_waitqueue_head(&is->wq);
	}

	up(&state->sem);

	DPRINTK("ep93xx_audio_open -- EXIT\n");

	/*ac97 controller and codec startup*/
	err = state->startup(substream);
	if (!err)
		goto out;
 out:
	return err;

	
}

/*
 *free the HW dma channel
 *free the HW dma buffer
 *free the Hw dma decrotion using function :kfree
 *shut down the ac97 codec and  Ac97 controller
 */
static int ep93xx_ac97_pcm_close(snd_pcm_substream_t *substream)
{
	audio_state_t *state = substream->private_data;
	int ret;

	DPRINTK("ep93xx_ac97_pcm_close----ENTER\n");

	ret = audio_release(substream);
	if(ret!=0)
		printk("%s error %x",__FUNCTION__,ret);

	/*ac97 controller and codec shutdown*/
	state->shutdown(substream);

	DPRINTK("ep93xx_ac97_pcm_close----Over\n");
	return 0;
}


/*
 *map the driver memory space to user memory space
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
static int
ep93xx_ac97_pcm_mmap(snd_pcm_substream_t *substream, struct vm_area_struct *vma)
{
		
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	snd_pcm_runtime_t *runtime = substream->runtime;
	//snd_pcm_lib_mmap_iomem
	return dma_mmap_writecombine(substream->pcm->card->dev, vma,
				     runtime->dma_area,
				     runtime->dma_addr,
				     runtime->dma_bytes);

	#else
	return 0;
	#endif
}
#endif

/* */
static int ep93xx_ac97_pcm_copy(snd_pcm_substream_t * substream,int channel, snd_pcm_uframes_t pos,void __user *src, snd_pcm_uframes_t count)
{

	audio_state_t *state = (audio_state_t *)substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;

	int tocount=frames_to_bytes(runtime,count);
	/*unsigned long bpos=frames_to_bytes(runtime,pos);*/
	int ret=0;


/*	ret=copy_n_convert(state,substream->dma_buffer.area,(const char*)src+bpos,tocount);
	printk("ret=%x\n",ret);
	ret=count;
	return ret;
*/
	size_t dma_csize = runtime->dma_bytes / runtime->channels;
	char *hwbuf ;
	if(channel==-1){
		//hwbuf	= runtime->dma_area  + samples_to_bytes(runtime, pos)*2*2;
		hwbuf	= runtime->dma_area  + scale_to_expanded_samples( state,samples_to_bytes(runtime, pos)*2,substream->stream);
		DPRINTK("copy frames=%x,%x,pos=%x,bpos=%x channel=%x\n ",count,tocount,pos,samples_to_bytes(runtime, pos),channel);
	//	return -EIO;
	}
	else{
		hwbuf	= runtime->dma_area + (channel * dma_csize) + samples_to_bytes(runtime, pos);
		DPRINTK("copy frames=%x,,pos=%x,bpos=%x channel=%x\n ",count,pos,(channel * dma_csize) + samples_to_bytes(runtime, pos),channel);
	}
	ret=copy_n_convert(state,hwbuf,(const char*)src,scale_to_expanded_samples(state,tocount,substream->stream));
	//ret=copy_n_convert(state,hwbuf,(const char*)src,tocount*2);
//	printk("ret=%x\n",ret);
	ret=0;//count;
	return ret;
}

int Copy_count;
/* */
static int ep93xx_ac97_pcm_capture_copy(snd_pcm_substream_t * substream,int channel, snd_pcm_uframes_t pos,void __user *src, snd_pcm_uframes_t count)
{
	audio_state_t *state 		= (audio_state_t *)(substream->private_data);
	snd_pcm_runtime_t *runtime 	= substream->runtime;
	audio_stream_t *s 			= state->input_stream;

	char *hwbuf = runtime->dma_area + scale_to_expanded_samples(state,frames_to_bytes(runtime, pos),substream->stream);//frames_to_bytes(runtime, pos);
	

	audio_buf_t *buf ;//= s->buf;
        /*int tocount=frames_to_bytes(runtime,count);*/
        /*unsigned long bpos=frames_to_bytes(runtime,pos);*/

     DPRINTK("copy frames=%x,%x,pos=%x,bpos=%x channel=%x\n ",count,tocount,pos,samples_to_bytes(runtime, pos),channel);


/*
	up(&buf->sem);

	printk("copy:%x\n",s->buf->size);
	if (down_interruptible(&buf->sem)){
		printk("sem eror\n");
		return -1;
	}
	printk("copy1\n");
	//if (copy_to_user((char *)src, hwbuf, frames_to_bytes(runtime, count)))
	//{
	//	printk("copy to user eror\n");
	//	up(&buf->sem);
	//	return -EFAULT;
	//}
	printk("copy2\n");

*/
	if(copy_to_user_with_conversion(state,(char *)src,hwbuf, frames_to_bytes(runtime, count))==0)
	//if (copy_to_user((char *)src, hwbuf, frames_to_bytes(runtime, count)))
	{
		DPRINTK("copy to user eror\n");
		/*up(&buf->sem);*/
		return -EFAULT;
	}
    if(pos==0)	Copy_count=0;

    if(Copy_count<=frames_to_bytes(runtime,runtime->buffer_size)){
	//stoptmp=1;
	
	DPRINTK("WriteEnd %x\n",WriteEnd);
	buf=(audio_buf_t *)Stack_buf_id[WriteEnd];
	WriteEnd	=	(WriteEnd	+1)%(s->nbfrags);
	DPRINTK("1\n");
        if (down_interruptible(&buf->sem)){
                DPRINTK("sem eror\n");
                return -1;
        }


	ep93xx_dma_add_buffer( s->dmahandles[0], 	// dma instance 		
						   (unsigned int)buf->dma_addr, // source 		
						   0, 					// dest 				
						   buf->size,	//s->fragsize,     size 	
						   0,					// is the last chunk? 	
						   (unsigned int) buf );//  buf id	

	DPRINTK("2\n");
	Copy_count=Copy_count+buf->reportedsize;
	DPRINTK("copy %x %x\n",buf->size,Copy_count);
	
   	NEXT_BUF(s, buf);

	DPRINTK("3\n");
	up(&buf->sem);
	//stoptmp=0;
//	up(&sLoopsem);
	DPRINTK("4\n");
    }
	return 0;
	
}

static snd_pcm_ops_t ep93xx_ac97_pcm_ops = {
	.open		= ep93xx_ac97_pcm_open,
	.close		= ep93xx_ac97_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= ep93xx_ac97_pcm_hw_params,
	.hw_free	= ep93xx_ac97_pcm_hw_free,
	.prepare	= ep93xx_ac97_pcm_prepare,
	.trigger	= ep93xx_ac97_pcm_trigger,
	.pointer	= ep93xx_ac97_pcm_pointer,
	.copy 		= ep93xx_ac97_pcm_copy,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	.mmap		= ep93xx_ac97_pcm_mmap,
#endif

};

static snd_pcm_ops_t ep93xx_ac97_capture_pcm_ops = {
	.open		= ep93xx_ac97_pcm_open,
	.close		= ep93xx_ac97_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= ep93xx_ac97_pcm_hw_params,
	.hw_free	= ep93xx_ac97_pcm_hw_free,
	.prepare	= ep93xx_ac97_pcm_prepare,
	.trigger	= ep93xx_ac97_pcm_trigger,
	.pointer	= ep93xx_ac97_pcm_pointer,
	.copy 		= ep93xx_ac97_pcm_capture_copy,
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	.mmap		= ep93xx_ac97_pcm_mmap,
#endif

};

/*
 *-----------------------------------------------------------------------------------*
 */
/*
static int ep93xx_ac97_pcm_preallocate_dma_buffer(snd_pcm_t *pcm, int stream)
{
	snd_pcm_substream_t *substream = pcm->streams[stream].substream;
	struct snd_dma_buffer *buf = &substream->dma_buffer;
	size_t size = ep93xx_ac97_pcm_hardware.buffer_bytes_max;

	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	buf->dev.type = SNDRV_DMA_TYPE_DEV;
	buf->dev.dev = pcm->card->dev;
	#endif
	
	buf->private_data = NULL;
	buf->area = dma_alloc_writecombine(pcm->card->dev, size,
					   &buf->addr, GFP_KERNEL);
	if (!buf->area){
		DPRINTK("%s buf->area=%x,size=%x\n,buf->addr=%x",__FUNCTION__,buf->area,size,buf->addr);
		return -ENOMEM;
	}

	buf->bytes = size;
	return 0;
}

static void ep93xx_ac97_pcm_free_dma_buffers(snd_pcm_t *pcm)
{
	snd_pcm_substream_t *substream;
	struct snd_dma_buffer *buf;
	int stream;

	for (stream = 0; stream < 2; stream++) {
		substream = pcm->streams[stream].substream;
		if (!substream)
			continue;
		buf = &substream->dma_buffer;
		if (!buf->area)
			continue;
		dma_free_writecombine(pcm->card->dev, buf->bytes,
				      buf->area, buf->addr);
		buf->area = NULL;
	}
}
*/




static u64 ep93xx_ac97_pcm_dmamask = 0xffffffff;

int ep93xx_ac97_pcm_new(snd_card_t *card, audio_state_t *state, snd_pcm_t **rpcm)
{
	snd_pcm_t *pcm;
	int play = state->output_stream? 1 : 0;/*SNDRV_PCM_STREAM_PLAYBACK*/
	int capt = state->input_stream ? 1 : 0;/*SNDRV_PCM_STREAM_CAPTURE*/
	int ret;

	Copy_count	=	0;
	ReadEnd	=	0;
	WriteEnd	=	0;
	DPRINTK("ep93xx_ac97_pcm_new----ENTER\n");

	/*
	 *Register the new pcm device interface
	 */
	ret = snd_pcm_new(card, "EP93xx-AC97-PCM", 0, play, capt, &pcm);
	if (ret){
		DPRINTK("%s--%x:card=%x,play=%x,capt=%x,&pcm=%x\n",__FUNCTION__,ret,card,play,capt,pcm);
		goto out;
	}

	/*
	 *Allocate the pcm(DMA) memory  
	 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
#else
/*
	ret=snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_CONTINUOUS,
						snd_dma_continuous_data(GFP_KERNEL), 128*1024, 128*1024);
*/
	ret=snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,0, 2*128*1024, 2*128*1024);
#endif

	DPRINTK("err=%x---1add=%x,1area=%x,1bytes=%x,2add=%x,2area=%x,2bytes=%x\n",ret,
						pcm->streams[0].substream->dma_buffer.addr,
						pcm->streams[0].substream->dma_buffer.area,
						pcm->streams[0].substream->dma_buffer.bytes,
						pcm->streams[1].substream->dma_buffer.addr,
						pcm->streams[1].substream->dma_buffer.area,
						pcm->streams[1].substream->dma_buffer.bytes);	


	pcm->private_data = state;

/*
	the code for using PIO loop way to acp97
	sema_init(&(sLoopsem),0);
	ret = kernel_thread(TxRxloop,0, 0);
	DPRINTK("kernel_thread init ret=%x\n",ret);
*/	
	 sema_init(&(sLoopsem),1);
	/*seem to free the pcm data struct-->self dma buffer*/
	#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	pcm->private_free = ep93xx_ac97_pcm_free_dma_buffers;
	#else
	pcm->private_free = (void *)snd_pcm_lib_preallocate_free_for_all;
	#endif

	
	/*in the ep93 arch file,have similar setting*/
	if (!card->dev->dma_mask)
		card->dev->dma_mask = &ep93xx_ac97_pcm_dmamask;
	if (!card->dev->coherent_dma_mask)
		card->dev->coherent_dma_mask = 0xffffffff;


	/*
	 *alsa pcm ops setting for SNDRV_PCM_STREAM_PLAYBACK
	 */
	if (play) {
		int stream = SNDRV_PCM_STREAM_PLAYBACK;
		snd_pcm_set_ops(pcm, stream, &ep93xx_ac97_pcm_ops);
		
		#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
		ret = ep93xx_ac97_pcm_preallocate_dma_buffer(pcm, stream);
		if (ret){
			DPRINTK("%s--%x:pcm=%x,stream=%x\n",__FUNCTION__,ret,pcm,stream);
			goto out;
		}
		#endif
	}

	/*
	 *alsa pcm ops setting for SNDRV_PCM_STREAM_CAPTURE
	 */	
	if (capt) {
		int stream = SNDRV_PCM_STREAM_CAPTURE;
		snd_pcm_set_ops(pcm, stream, &ep93xx_ac97_capture_pcm_ops);

		#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
		ret = ep93xx_ac97_pcm_preallocate_dma_buffer(pcm, stream);
		if (ret){
			DPRINTK("%s--%x:pcm=%x,stream=%x\n",__FUNCTION__,ret,pcm,stream);;
			goto out;
		}
		#endif
	}

	if (rpcm)
		*rpcm = pcm;
	
	ret = 0;
	
	DPRINTK("ep93xx_ac97_pcm_new----EXIT\n");
	
 out:
	return ret;
}

EXPORT_SYMBOL(ep93xx_ac97_pcm_new);

MODULE_AUTHOR("Shrek");
MODULE_DESCRIPTION("Cirrus Ep9xxx AC97 PCM module");
MODULE_LICENSE("GPL");
